"""Localization for the Hijri month and day names."""

from typing import ClassVar, Dict, List, Type

_locale_map: Dict[str, Type["Locale"]] = {}


def get_locale(name: str) -> "Locale":
    """Return an appropriate :obj:`Locale` corresponding to a locale name.

    Args:
        name: name of the locale.
    """

    language_tag = name.lower()[:2]
    locale_cls = _locale_map.get(language_tag)

    if locale_cls is None:
        raise ValueError(f"Unsupported language: {language_tag}")

    return locale_cls()


class Locale:
    """A Locale object represents locale-specific data and functionality."""

    language_tag: ClassVar[str]
    month_names: ClassVar[List[str]] = []
    gregorian_month_names: ClassVar[List[str]] = []
    day_names: ClassVar[List[str]] = []
    notation: ClassVar[str]
    gregorian_notation: ClassVar[str]

    def __init_subclass__(cls) -> None:
        if cls.language_tag in _locale_map:
            raise LookupError(f"Duplicated language tag: {cls.language_tag}")
        _locale_map[cls.language_tag] = cls

    def month_name(self, month: int) -> str:
        """Return the month name for a specified Hijri month of the year.

        Args:
            month: month of year, in range 1-12.
        """

        return self.month_names[month - 1]

    def gregorian_month_name(self, month: int) -> str:
        """Return the month name for a specified Gregorian month of the year.

        Args:
            month: month of year, in range 1-12.
        """

        return self.gregorian_month_names[month - 1]

    def day_name(self, day: int) -> str:
        """Return the day name for a specified day of the week.

        Args:
            day: day of week, where Monday is 1 and Sunday is 7.
        """

        return self.day_names[day - 1]


class EnglishLocale(Locale):
    """An English Locale object represents English locale-specific data and
    functionality.
    """

    language_tag = "en"
    month_names = [
        "Muharram",
        "Safar",
        "Rabi’ al-Awwal",
        "Rabi’ al-Thani",
        "Jumada al-Ula",
        "Jumada al-Akhirah",
        "Rajab",
        "Sha’ban",
        "Ramadhan",
        "Shawwal",
        "Dhu al-Qi’dah",
        "Dhu al-Hijjah",
    ]
    gregorian_month_names = [
        "January",
        "February",
        "March",
        "April",
        "May",
        "June",
        "July",
        "August",
        "September",
        "October",
        "November",
        "December",
    ]
    day_names = [
        "Monday",
        "Tuesday",
        "Wednesday",
        "Thursday",
        "Friday",
        "Saturday",
        "Sunday",
    ]
    notation = "AH"
    gregorian_notation = "CE"


class ArabicLocale(Locale):
    """An Arabic Locale object represents Arabic locale-specific data and
    functionality.
    """

    language_tag = "ar"
    month_names = [
        "محرم",
        "صفر",
        "ربيع الأول",
        "ربيع الثاني",
        "جمادى الأولى",
        "جمادى الآخرة",
        "رجب",
        "شعبان",
        "رمضان",
        "شوال",
        "ذو القعدة",
        "ذو الحجة",
    ]
    gregorian_month_names = [
        "يناير",
        "فبراير",
        "مارس",
        "أبريل",
        "مايو",
        "يونيو",
        "يوليو",
        "أغسطس",
        "سبتمبر",
        "أكتوبر",
        "نوفمبر",
        "ديسمبر",
    ]
    day_names = [
        "الإثنين",
        "الثلاثاء",
        "الأربعاء",
        "الخميس",
        "الجمعة",
        "السبت",
        "الأحد",
    ]
    notation = "هـ"
    gregorian_notation = "م"


class BengaliLocale(Locale):
    """A Bengali Locale object represents Bengali locale-specific data and
    functionality.
    """

    language_tag = "bn"
    month_names = [
        "মুহাররম",
        "সফর",
        "রবিউল আউয়াল",
        "রবিউস সানী",
        "জুমাদাল উলা",
        "জুমাদাস সানী",
        "রজব",
        "শাবান",
        "রমজান",
        "শাওয়াল",
        "জিলক্বদ",
        "জিলহজ",
    ]
    gregorian_month_names = [
        "জানুয়ারি",
        "ফেব্রুয়ারি",
        "মার্চ",
        "এপ্রিল",
        "মে",
        "জুন",
        "জুলাই",
        "আগস্ট",
        "সেপ্টেম্বর",
        "অক্টোবর",
        "নভেম্বর",
        "ডিসেম্বর",
    ]
    day_names = [
        "সোমবার",
        "মঙ্গলবার",
        "বুধবার",
        "বৃহস্পতিবার",
        "শুক্রবার",
        "শনিবার",
        "রবিবার",
    ]
    notation = "হিজরি"
    gregorian_notation = "খ্রিস্টাব্দ"



from hijridate import helpers


def test_julian_to_ordinal():
    assert helpers.jdn_to_ordinal(2447977) == 726552


def test_ordinal_to_julian():
    assert helpers.ordinal_to_jdn(726552) == 2447977


def test_julian_to_reduced_julian():
    assert helpers.jdn_to_rjd(2456087) == 56087


def test_reduced_julian_to_julian():
    assert helpers.rjd_to_jdn(56087) == 2456087



from datetime import date

import pytest

from hijridate.convert import Gregorian, Hijri


def test_importing_at_init_module():
    from hijridate import Gregorian, Hijri

    assert Hijri(1410, 8, 13)
    assert Gregorian(1990, 3, 10)


@pytest.fixture(scope="class")
def hijri_obj(request):
    request.cls.hijri_obj = Hijri(1410, 8, 13)


@pytest.mark.usefixtures("hijri_obj")
class TestHijri:
    def test_representation(self):
        assert self.hijri_obj.__repr__() == "Hijri(1410, 8, 13)"

    def test_string_representation(self):
        assert self.hijri_obj.__str__() == "1410-08-13"

    def test_hash(self):
        assert self.hijri_obj.__hash__() == hash(("Hijri", 1410, 8, 13))

    @pytest.mark.parametrize(
        "test_input", ["__gt__", "__ge__", "__lt__", "__le__"]
    )
    def test_comparison_notimplemented(self, test_input):
        assert (
            getattr(self.hijri_obj, test_input)("1410-08-13") == NotImplemented
        )

    def test_equality(self):
        assert self.hijri_obj == Hijri(1410, 8, 13)
        assert self.hijri_obj != Hijri(1410, 8, 14)
        assert self.hijri_obj != "1410-08-13"

    def test_ordering(self):
        assert self.hijri_obj > Hijri(1410, 8, 12)
        assert self.hijri_obj >= Hijri(1410, 8, 13)
        assert self.hijri_obj < Hijri(1410, 8, 14)
        assert self.hijri_obj <= Hijri(1410, 8, 13)

    def test_fromisoformat(self):
        assert Hijri.fromisoformat("1410-08-13") == self.hijri_obj

    def test_today(self):
        assert Hijri.today().to_gregorian() == Gregorian.today()

    def test_year(self):
        assert self.hijri_obj.year == 1410

    def test_month(self):
        assert self.hijri_obj.month == 8

    def test_day(self):
        assert self.hijri_obj.day == 13

    def test_datetuple(self):
        assert self.hijri_obj.datetuple() == (1410, 8, 13)

    def test_isoformat(self):
        assert self.hijri_obj.isoformat() == "1410-08-13"

    def test_dmyformat(self):
        assert self.hijri_obj.dmyformat() == "13/08/1410"
        assert self.hijri_obj.dmyformat(padding=False) == "13/8/1410"
        assert self.hijri_obj.dmyformat(separator=".") == "13.08.1410"

    def test_month_length(self):
        assert self.hijri_obj.month_length() == 29

    def test_month_name(self):
        assert self.hijri_obj.month_name() == "Sha’ban"
        assert self.hijri_obj.month_name("en") == "Sha’ban"
        assert self.hijri_obj.month_name("en-US") == "Sha’ban"

    def test_weekday(self):
        assert self.hijri_obj.weekday() == 5

    def test_iso_weekday(self):
        assert self.hijri_obj.isoweekday() == 6

    def test_day_name(self):
        assert self.hijri_obj.day_name() == "Saturday"
        assert self.hijri_obj.day_name("en") == "Saturday"
        assert self.hijri_obj.day_name("en-US") == "Saturday"

    def test_notation(self):
        assert self.hijri_obj.notation() == "AH"
        assert self.hijri_obj.notation("en") == "AH"
        assert self.hijri_obj.notation("en-US") == "AH"

    def test_to_julian(self):
        assert self.hijri_obj.to_julian() == 2447961

    def test_to_gregorian(self):
        assert self.hijri_obj.to_gregorian().datetuple() == (1990, 3, 10)

    def test_month_index(self):
        assert self.hijri_obj._month_index() == 811

    @pytest.mark.parametrize(
        "test_input", [(1410, 9, 30), (1356, 1, 1), (1500, 12, 30)]
    )
    def test_valid_date(self, test_input):
        year, month, day = test_input
        try:
            Hijri(year, month, day, validate=False)._check_date()
        except (ValueError, OverflowError):
            pytest.fail()

    @pytest.mark.parametrize(
        "test_input, expected",
        [
            ((37, 12, 30), "date out of range"),
            ((1342, 12, 29), "date out of range"),
            ((1501, 1, 1), "date out of range"),
        ],
    )
    def test_invalid_year(self, test_input, expected):
        with pytest.raises(OverflowError) as e:
            Hijri(*test_input, validate=False)._check_date()
        assert str(e.value) == expected

    @pytest.mark.parametrize(
        "test_input, expected",
        [
            ((1410, 0, 1), "month must be in 1..12"),
            ((1410, 13, 1), "month must be in 1..12"),
            ((1410, 8, 30), "day must be in 1..29 for month"),
        ],
    )
    def test_invalid_day_or_month(self, test_input, expected):
        with pytest.raises(ValueError) as e:
            Hijri(*test_input, validate=False)._check_date()
        assert str(e.value) == expected


@pytest.fixture(scope="class")
def gregorian_obj(request):
    request.cls.gregorian_obj = Gregorian(1990, 3, 10)


@pytest.mark.usefixtures("gregorian_obj")
class TestGregorian:
    def test_fromdate(self):
        test_date = date(2014, 12, 28)
        assert Gregorian.fromdate(test_date).datetuple() == (2014, 12, 28)

    def test_datetuple(self):
        assert self.gregorian_obj.datetuple() == (1990, 3, 10)

    def test_dmyformat(self):
        assert self.gregorian_obj.dmyformat() == "10/03/1990"
        assert self.gregorian_obj.dmyformat(padding=False) == "10/3/1990"
        assert self.gregorian_obj.dmyformat(separator=".") == "10.03.1990"

    def test_month_name(self):
        assert self.gregorian_obj.month_name() == "March"
        assert self.gregorian_obj.month_name("en") == "March"
        assert self.gregorian_obj.month_name("en-US") == "March"

    def test_day_name(self):
        assert self.gregorian_obj.day_name() == "Saturday"
        assert self.gregorian_obj.day_name("en") == "Saturday"
        assert self.gregorian_obj.day_name("en-US") == "Saturday"

    def test_notation(self):
        assert self.gregorian_obj.notation() == "CE"
        assert self.gregorian_obj.notation("en") == "CE"
        assert self.gregorian_obj.notation("en-US") == "CE"

    def test_to_julian(self):
        assert self.gregorian_obj.to_julian() == 2447961

    def test_to_hijri(self):
        assert self.gregorian_obj.to_hijri().datetuple() == (1410, 8, 13)

    @pytest.mark.parametrize(
        "test_input", [(1990, 3, 10), (1924, 8, 1), (2077, 11, 16)]
    )
    def test_valid_range(self, test_input):
        year, month, day = test_input
        try:
            Gregorian(year, month, day)._check_range()
        except OverflowError:
            pytest.fail()

    @pytest.mark.parametrize(
        "test_input, expected",
        [
            ((1924, 7, 31), "date out of range"),
            ((2077, 11, 17), "date out of range"),
        ],
    )
    def test_invalid_range(self, test_input, expected):
        with pytest.raises(OverflowError) as e:
            Gregorian(*test_input)._check_range()
        assert str(e.value) == expected



"""Main module of the hijridate package."""

import datetime
from bisect import bisect
from typing import Tuple

from hijridate import helpers, locales, ummalqura


class Hijri:
    """A Hijri object represents a date (year, month and day) in lunar Hijri
    calendar.
    """

    __slots__ = "_year", "_month", "_day"

    def __init__(self, year: int, month: int, day: int, validate: bool = True):
        """
        Args:
            year: Hijri year.
            month: Hijri month.
            day: Hijri day.
            validate: Whether to validate Hijri input or not. It's recommended
                to keep the default for accurate conversion.

        Raises:
            OverflowError: When ``year`` is out of supported Hijri range.
            ValueError: When ``month`` is not within the range of `1..12`.
            ValueError: When ``day`` is not within the range of
                `1..month_length` for month.
        """

        self._year = year
        self._month = month
        self._day = day

        if validate:
            self._check_date()

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}({self._year}, {self._month}, {self._day})"

    def __str__(self) -> str:
        return self.isoformat()

    def __hash__(self) -> int:
        return hash(("Hijri", self._year, self._month, self._day))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Hijri):
            return NotImplemented
        return self._compare(other) == 0

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, Hijri):
            return NotImplemented
        return self._compare(other) > 0

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, Hijri):
            return NotImplemented
        return self._compare(other) >= 0

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, Hijri):
            return NotImplemented
        return self._compare(other) < 0

    def __le__(self, other: object) -> bool:
        if not isinstance(other, Hijri):
            return NotImplemented
        return self._compare(other) <= 0

    def _compare(self, other: "Hijri") -> int:
        self_date = self.datetuple()
        other_date = other.datetuple()
        return (
            0
            if self_date == other_date
            else 1
            if self_date > other_date
            else -1
        )

    @classmethod
    def fromisoformat(cls, date_string: str) -> "Hijri":
        """Construct Hijri object from an ISO formatted Hijri date
        ``YYYY-MM-DD``.

        Args:
            date_string: Hijri date in ISO format ``YYYY-MM-DD``.
        """

        year = int(date_string[0:4])
        month = int(date_string[5:7])
        day = int(date_string[8:10])
        return cls(year, month, day)

    @classmethod
    def today(cls) -> "Hijri":
        """Construct Hijri object from today's date."""
        return Gregorian.today().to_hijri()

    @property
    def year(self) -> int:
        """Return year as an integer."""
        return self._year

    @property
    def month(self) -> int:
        """Return month as an integer."""
        return self._month

    @property
    def day(self) -> int:
        """Return day as an integer."""
        return self._day

    def datetuple(self) -> Tuple[int, int, int]:
        """Return date as a tuple of (year, month, day)."""
        return self._year, self._month, self._day

    def isoformat(self) -> str:
        """Return date in ISO format i.e. ``YYYY-MM-DD``."""
        return f"{self._year:04}-{self._month:02}-{self._day:02}"

    def dmyformat(self, separator: str = "/", padding: bool = True) -> str:
        """Return date in day-month-year format (``DD/MM/YYYY`` by default).

        Args:
            separator: String that separates the day, month, and year values.
            padding: Whether to add a leading zero as a padding character to
                fill day and month values when less than 10.
        """

        day = f"{self._day:02}" if padding else self._day
        month = f"{self._month:02}" if padding else self._month
        return f"{day}{separator}{month}{separator}{self._year}"

    def month_length(self) -> int:
        """Return number of days in month."""
        month_starts = ummalqura.MONTH_STARTS
        index = self._month_index()
        length = month_starts[index + 1] - month_starts[index]
        return length

    def month_name(self, language: str = "en") -> str:
        """Return month name.

        Args:
            language: Language tag for localized month name. Full locale name
                can be used, e.g. ``en-US`` or ``en_US.UTF-8``. Supported
                languages are ``en``, ``ar`` and ``bn``.
        """

        return locales.get_locale(language).month_name(self._month)

    def weekday(self) -> int:
        """Return day of week, where Monday is 0 and Sunday is 6."""
        jdn = self.to_julian()
        return int(jdn % 7)

    def isoweekday(self) -> int:
        """Return day of week, where Monday is 1 and Sunday is 7."""
        jdn = self.to_julian()
        return int(jdn % 7) + 1

    def day_name(self, language: str = "en") -> str:
        """Return day name.

        Args:
            language: Language tag for localized month name. Full locale name
                can be used, e.g. ``en-US`` or ``en_US.UTF-8``. Supported
                languages are ``en``, ``ar`` and ``bn``.
        """

        return locales.get_locale(language).day_name(self.isoweekday())

    @staticmethod
    def notation(language: str = "en") -> str:
        """Return calendar era notation.

        Args:
            language: Language tag for localized month name. Full locale name
                can be used, e.g. ``en-US`` or ``en_US.UTF-8``. Supported
                languages are ``en``, ``ar`` and ``bn``.
        """

        return locales.get_locale(language).notation

    def to_julian(self) -> int:
        """Return corresponding Julian day number (JDN)."""
        month_starts = ummalqura.MONTH_STARTS
        index = self._month_index()
        rjd = month_starts[index] + self._day - 1
        jdn = helpers.rjd_to_jdn(rjd)
        return jdn

    def to_gregorian(self) -> "Gregorian":
        """Return Gregorian object for the corresponding Hijri date."""

        jdn = self.to_julian()
        ordinal = helpers.jdn_to_ordinal(jdn)
        return Gregorian.fromordinal(ordinal)

    def _check_date(self) -> None:
        """Check date values if within valid range."""

        # check year
        min_year, max_year = [d[0] for d in ummalqura.HIJRI_RANGE]
        if not min_year <= self.year <= max_year:
            raise OverflowError("date out of range")
        # check month
        max_months = 12
        if not 1 <= self.month <= max_months:
            raise ValueError(f"month must be in 1..{max_months}")
        # check day
        month_length = self.month_length()
        if not 1 <= self.day <= month_length:
            raise ValueError(f"day must be in 1..{month_length} for month")

    def _month_index(self) -> int:
        """Return month’s index in ummalqura month starts"""
        prior_months = ((self.year - 1) * 12) + self.month - 1
        index = prior_months - ummalqura.HIJRI_OFFSET
        return index


class Gregorian(datetime.date):
    """A Gregorian object represents a date (year, month and day) in Gregorian
    calendar.
    """

    __slots__ = ()

    @classmethod
    def fromdate(cls, date_object: datetime.date) -> "Gregorian":
        """Construct Gregorian object from a Python date object.

        Args:
            date_object: Python date object.
        """

        year, month, day = date_object.timetuple()[:3]
        return cls(year, month, day)

    def datetuple(self) -> Tuple[int, int, int]:
        """Return date as a tuple of (year, month, day)."""
        return self.year, self.month, self.day

    def dmyformat(self, separator: str = "/", padding: bool = True) -> str:
        """Return date in day-month-year format (``DD/MM/YYYY`` by default).

        Args:
            separator: String that separates the day, month, and year values.
            padding: Whether to add a leading zero as a padding character to
                fill day and month values when less than 10.
        """

        day = f"{self.day:02}" if padding else self.day
        month = f"{self.month:02}" if padding else self.month
        return f"{day}{separator}{month}{separator}{self.year}"

    def month_name(self, language: str = "en") -> str:
        """Return month name.

        Args:
            language: Language tag for localized month name. Full locale name
                can be used, e.g. ``en-US`` or ``en_US.UTF-8``. Supported
                languages are ``en``, ``ar`` and ``bn``.
        """

        return locales.get_locale(language).gregorian_month_name(self.month)

    def day_name(self, language: str = "en") -> str:
        """Return day name.

        Args:
            language: Language tag for localized month name. Full locale name
                can be used, e.g. ``en-US`` or ``en_US.UTF-8``. Supported
                languages are ``en``, ``ar`` and ``bn``.
        """

        return locales.get_locale(language).day_name(self.isoweekday())

    @staticmethod
    def notation(language: str = "en") -> str:
        """Return calendar era notation.

        Args:
            language: Language tag for localized month name. Full locale name
                can be used, e.g. ``en-US`` or ``en_US.UTF-8``. Supported
                languages are ``en``, ``ar`` and ``bn``.
        """

        return locales.get_locale(language).gregorian_notation

    def to_julian(self) -> int:
        """Return corresponding Julian day number (JDN)."""
        ordinal = self.toordinal()
        jdn = helpers.ordinal_to_jdn(ordinal)
        return jdn

    def to_hijri(self) -> Hijri:
        """Return Hijri object for the corresponding Gregorian date.

        Raises:
            OverflowError: When date is out of supported Gregorian range.
        """

        self._check_range()
        jdn = self.to_julian()
        rjd = helpers.jdn_to_rjd(jdn)
        month_starts = ummalqura.MONTH_STARTS
        index = bisect(month_starts, rjd) - 1
        months = index + ummalqura.HIJRI_OFFSET
        years = int(months / 12)
        year = years + 1
        month = months - (years * 12) + 1
        day = rjd - month_starts[index] + 1
        return Hijri(year, month, day, validate=False)

    def _check_range(self) -> None:
        """Check if Gregorian date is within valid range."""
        min_date, max_date = ummalqura.GREGORIAN_RANGE
        if not min_date <= (self.year, self.month, self.day) <= max_date:
            raise OverflowError("date out of range")



import json
from pathlib import Path

import pytest

from hijridate import Gregorian, Hijri


def load_params_from_json():
    data = []
    files = Path(__file__).parent.joinpath("fixtures").glob("*.json")
    for file in files:
        file_content = json.loads(file.read_text())
        # skip test data when the last element is False
        file_data = [
            [tuple(i) for i in x[0]] for x in file_content if x[-1] is not False
        ]
        data += file_data
    params = [tuple(x) for x in data]
    params_reversed = [tuple(x[::-1]) for x in data]
    return params, params_reversed


hijri_gregorian_params, gregorian_hijri_params = load_params_from_json()


@pytest.mark.parametrize("test_input, expected", hijri_gregorian_params)
def test_convert_hijri_to_gregorian(test_input, expected):
    year, month, day = test_input
    hijri = Hijri(year, month, day, validate=False)
    converted = hijri.to_gregorian().datetuple()
    assert converted == expected


@pytest.mark.parametrize("test_input, expected", gregorian_hijri_params)
def test_convert_gregorian_to_hijri(test_input, expected):
    year, month, day = test_input
    gregorian = Gregorian(year, month, day)
    converted = gregorian.to_hijri().datetuple()
    assert converted == expected
